// Teapot 3 - the Utah teapot, cup, plate and spoon tesselated on the GPU.
// By Ingemar Ragnemalm 2013, using something as close to the original data as possible,
// but with some fixes for the most disturbing problems.
// Based on my "glutgears3" demo, which was based on the "glutgears" demo.
// You may note that the usual bug in a certain place (the knob) is fixed.

// This version saves some of the generated models as OBJ files.
// New version 2022 for GLUGG 1.3.

// Tesselation density
// This is nice: #define step 0.1
// This is rough: #define step 0.5
#define kStep 0.05

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
// uses framework Cocoa
// uses framework OpenGL
#define MAIN
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils4.h"
#include "teaset-data.h"
#include "glugg.h"
#include "saveobj.h"

#ifndef M_PI
#define M_PI 3.14159265
#endif

static GLfloat view_rotx = 20.0, view_roty = 30.0, view_rotz = 0.0;
static GLfloat angle = 0.0;

// The teaset objects and lengths
gluggModel vao[12];

GLuint program; // Shader
mat4 projectionMatrix;

static GLfloat pos[4] = { 5.0, 5.0, 10.0, 0.0 };
static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };

static GLfloat colors[12][4] = {
									{ 0.8, 0.1, 0.0, 1.0 }, // Red
									{ 0.0, 0.8, 0.2, 1.0 }, // Green
									{ 0.2, 0.2, 1.0, 1.0 }, // Blue
									{ 0.7, 0.7, 0.7, 1.0 }, // Gray
									{ 0.7, 0.2, 0.7, 1.0 }, // Purple/magenta
									{ 1.0, 1.0, 0.0, 1.0 }, // Bright yellow
									{ 1.0, 0.6, 0.0, 1.0 }, // Orange?
									{ 0.0, 1.0, 1.0, 1.0 }, // Bright cyan
									{ 1.0, 1.0, 1.0, 1.0 }, // White
									{ 1.0, 0.0, 0.0, 1.0 }, // Bright red
									{ 0.0, 1.0, 0.0, 1.0 }, // Bright green
									{ 0.9, 0.5, 0.6, 1.0 } // Pale red
									};


static void
draw(void)
{
	printError("pre display");

	angle += 2.0;
	mat4 rx, ry, rz, globalR, t, teapotRot, m;
	int i;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
// GLOBAL ROTATION
	t = T(0, 0, -40.0);
	rx = Rx(view_rotx*M_PI/180);
	ry = Ry(-view_roty*M_PI/180);
	rz = Rz(view_rotz*M_PI/180);
	globalR = Mult(rx, ry);
	globalR = Mult(t, globalR);
	globalR = Mult(globalR, rz);
	
	for (i=0; i < 12; i++)
	{
		t = T(-5.0 + 4.0*(i/3), -6.0 + 5.0*(i%3), 0.0);
		rx = Rx(-M_PI/2);
		teapotRot = Rz(angle*M_PI/180*(i+1)/3);
		m = Mult(globalR, t);
		m = Mult(m, rx);
		m = Mult(m, teapotRot);
		glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, m.m);
		glUniform4fv(glGetUniformLocation(program, "color"), 1, colors[i]);
		gluggDrawModel(vao[i], program);
	}

	glutSwapBuffers();
	
	printError("display");
}


/* new window size or exposure */
static void
reshape(int width, int height)
{
	GLfloat h = (GLfloat) height / (GLfloat) width;

	glViewport(0, 0, (GLint) width, (GLint) height);
	projectionMatrix = frustum(-1.0, 1.0, -h, h, 5.0, 60.0);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
}

static void
init(void)
{
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	
	glClearColor(1,1,1,1);
	
	program = loadShaders("tea.vert", "tea.frag");
	glUniform4fv(glGetUniformLocation(program, "pos"), 1, pos);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);
	printf("SHADERS LOADED\n");
		
	/* make the teapots */
	vao[0] = gluggBuildBezierPatchModel(teapotVertices, (int *)teapotIndices, 0, 32, kStep*2);
	vao[1] = gluggBuildBezierPatchModel(teapotVertices, (int *)teapotIndices, 0, 32, kStep*4);
	vao[2] = gluggBuildBezierPatchModel(teapotVertices, (int *)teapotIndices, 0, 32, kStep*10);
	/* make the teacups */
	vao[3] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 0, 16, kStep*2);
	vao[4] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 0, 16, kStep*4);
	vao[5] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 0, 16, kStep*10);
	/* make the plates */
	vao[6] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 16, 30, kStep*2);
	vao[7] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 16, 30, kStep*4);
	vao[8] = gluggBuildBezierPatchModel(teacupVertices, (int *)teacupIndices, 16, 30, kStep*10);
	/* make the tea spoons */
	vao[9] = gluggBuildBezierPatchModel(teaspoonVertices, (int *)teaspoonIndices, 0, 16, kStep*2);
	vao[10] = gluggBuildBezierPatchModel(teaspoonVertices, (int *)teaspoonIndices, 0, 16, kStep*4);
	vao[11] = gluggBuildBezierPatchModel(teaspoonVertices, (int *)teaspoonIndices, 0, 16, kStep*10);
	
	/* save the teapot */
	SaveBezierPatchModel(teapotVertices, (int *)teapotIndices, 0, 32, kStep*2, "teapot.obj");
	/* save the teacup */
	SaveBezierPatchModel(teacupVertices, (int *)teacupIndices, 0, 16, kStep, "teacup.obj");
	/* save the plate */
	SaveBezierPatchModel(teacupVertices, (int *)teacupIndices, 16, 30, kStep, "teaplate.obj");
	/* save the tea spoon */
	SaveBezierPatchModel(teaspoonVertices, (int *)teaspoonIndices, 0, 16, kStep*2, "teaspoon.obj");

	/* save the teapot body */
	SaveBezierPatchModel(teapotVertices, (int *)teapotAltIndices, 0, 24, kStep*2, "teapotbody.obj");
	/* save the teapot lid */
	SaveBezierPatchModel(teapotVertices, (int *)teapotIndices, 20, 28, kStep*2, "teapotlid.obj");
	
	projectionMatrix = frustum(-1.0, 1.0, -1.0, 1.0, 5.0, 60.0);
}


void
keys(unsigned char c, int x, int y)
{
	switch (c)
	{
		case 0x1b: exit(0);
		case 'a': view_roty += 5.0; break;
		case 'd': view_roty -= 5.0; break;
		case 'w': view_rotx += 5.0; break;
		case 's': view_rotx -= 5.0; break;
	}
}


int
main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(600, 600);
	glutInitContextVersion(3, 2);
	glutCreateWindow("CPU-tesselated Utah teapot, cup, plate and spoon");
	
	glutReshapeFunc(reshape);
	glutDisplayFunc(draw);
	glutKeyboardFunc(keys);
	glutRepeatingTimer(15);

	init();

	glutMainLoop();

	return 0;
}
